#!/usr/bin/env python3
# #
# @copyright Copyright (C) 2023 SpyroSoft Solutions S.A. All rights reserved.
# #

from pathlib import Path
from typing import List, Optional

from clang.cindex import Cursor, CursorKind, Index

from scargo.config import Config
from scargo.file_generators.base_gen import create_file_from_template
from scargo.file_generators.mock_utils.data_classes import (
    ArgumentDescriptor,
    HeaderDescriptor,
    MockClassDescriptor,
    MockFunctionDescriptor,
    MockNamespaceDescriptor,
)

MOCKS_DIR = "tests/mocks"


def generate_mocks(src_header: Path, src_dir: str, config: Config) -> bool:
    """
    Generates mock header and implementations for specified source headers.
    Creates directories and CMake lists where required.
    :param src_header Path to source directory or header
    :param src_dir source directory name (dependent on target)
    :param config
    """

    dst_header = get_mock_path(src_header, src_dir)

    # Skip generation if destination header exists
    if dst_header.is_file():
        return False

    template_and_output_paths = [
        ("mock/class_interface.h.j2", dst_header),
        ("mock/class_mock.h.j2", dst_header.with_name(f"mock_{dst_header.name}")),
        ("mock/class_mock.cpp.j2", dst_header.with_name(f"mock_{dst_header.stem}.cpp")),
    ]

    # Read classes and functions from the source header
    header_data = parse_file(src_header)

    for template_path, output_path in template_and_output_paths:
        create_file_from_template(
            template_path,
            output_path,
            template_params={"header": header_data},
            config=config,
        )

    return True


def get_mock_path(path: Path, src_dir: str) -> Path:
    parts: List[str] = list(path.parts)
    for i, part in enumerate(parts):
        if part == src_dir:
            parts[i] = MOCKS_DIR
            break
    return Path(*parts)


class ParamsExtractor:
    @staticmethod
    def extract_cxx_methods(cursor: Cursor) -> List[MockFunctionDescriptor]:
        descriptors = (
            ParamsExtractor._extract_method_params(descendant)
            for descendant in cursor.walk_preorder()
            if descendant.kind == CursorKind.CXX_METHOD
        )

        return [x for x in descriptors if x is not None]

    @staticmethod
    def extract_namespaces(
        cursor: Cursor, filename: str
    ) -> List[MockNamespaceDescriptor]:
        return [
            ParamsExtractor._extract_namespace_params(descendant)
            for descendant in cursor.walk_preorder()
            if descendant.kind == CursorKind.NAMESPACE
            and descendant.location.file == filename
        ]

    @staticmethod
    def extract_classes(cursor: Cursor, filename: str) -> List[MockClassDescriptor]:
        return [
            ParamsExtractor._extract_class_params(descendant)
            for descendant in cursor.walk_preorder()
            if descendant.kind == CursorKind.CLASS_DECL
            and descendant.location.file == filename
        ]

    @staticmethod
    def _extract_method_params(
        cursor: Cursor,
    ) -> Optional[MockFunctionDescriptor]:
        fun_ret_val = cursor.result_type.spelling
        fun_name = cursor.spelling
        fun_args = []
        specifiers = []

        if cursor.access_specifier.name != "PUBLIC":
            return None
        if cursor.is_virtual_method():
            specifiers.append("override")
        if cursor.is_const_method():
            specifiers.append("const")
        for i in cursor.walk_preorder():
            if i.kind == CursorKind.PARM_DECL:
                fun_args.append(ArgumentDescriptor(i.spelling, i.type.spelling))
        return MockFunctionDescriptor(
            fun_name, fun_ret_val, specifiers, arguments=fun_args
        )

    @staticmethod
    def _extract_namespace_params(cursor: Cursor) -> MockNamespaceDescriptor:
        return MockNamespaceDescriptor(cursor.spelling)

    @staticmethod
    def _extract_class_params(cursor: Cursor) -> MockClassDescriptor:
        cls = MockClassDescriptor(cursor.spelling, f"Mock{cursor.spelling}")
        cls.methods = ParamsExtractor.extract_cxx_methods(cursor)
        return cls


def parse_file(file_path: Path) -> HeaderDescriptor:
    """TODO documentation
    look here for available types which can be extracted
    https://clang.llvm.org/doxygen/group__CINDEX.html
    This function receives a path with file name as an argument and returns
    list of includes, i.e. every string that starts with "#include",
    dict of classes where keys are "public", "private" and "mock_class_name"
    and list of namespaces, i.e. strings that start with "namespace".
    """
    file_name = file_path.name

    hdr = HeaderDescriptor(file_name)

    idx = Index.create()
    translation_unit = idx.parse(str(file_path), ["-x", "c++"])

    hdr.namespaces = ParamsExtractor.extract_namespaces(
        translation_unit.cursor, str(file_path)
    )
    hdr.classes = ParamsExtractor.extract_classes(
        translation_unit.cursor, str(file_path)
    )
    return hdr



# #
# @copyright Copyright (C) 2023 SpyroSoft Solutions S.A. All rights reserved.
# #

import subprocess
import sys
from pathlib import Path
from typing import Optional

from scargo.config import Config
from scargo.config_utils import prepare_config
from scargo.logger import get_logger

logger = get_logger()


def scargo_flash(
    app: bool, fs: bool, flash_profile: str, port: Optional[str] = None
) -> None:
    config = prepare_config()
    target = config.project.target

    if port and target.family != "esp32":
        logger.error("--port option is only supported for esp32 projects.")
        sys.exit(1)
    if target.family == "esp32":
        flash_esp32(config, app=app, fs=fs, flash_profile=flash_profile, port=port)
    elif target.family == "stm32":
        flash_stm32(config, flash_profile)
    else:
        logger.error("Flash command not supported for this target yet.")


def flash_esp32(
    config: Config,
    app: bool,
    fs: bool,
    flash_profile: str = "Debug",
    port: Optional[str] = None,
) -> None:
    project_path = config.project_root
    out_dir = project_path / "build" / flash_profile
    target = config.project.target
    command = []
    try:
        if app:
            app_name = config.project.name
            app_path = out_dir / f"{app_name}.bin"
            command = [
                "parttool.py",
                "write_partition",
                "--partition-name=ota_0",
                f"--input={app_path}",
            ]
            if port:
                command.append(f"--port={port}")
            subprocess.check_call(command, cwd=project_path)
        elif fs:
            fs_path = Path("build") / "spiffs.bin"
            command = [
                "parttool.py",
                "write_partition",
                "--partition-name=spiffs",
                f"--input={fs_path}",
            ]
            if port:
                command.append(f"--port={port}")
            subprocess.check_call(command, cwd=project_path)
        else:
            command = ["esptool.py", "--chip", target.id, "write_flash", "@flash_args"]
            if port:
                command.append(f"--port={port}")
            subprocess.check_call(command, cwd=out_dir)
    except subprocess.CalledProcessError:
        logger.error("%s fail", command)


def flash_stm32(config: Config, flash_profile: str = "Debug") -> None:
    project_path = config.project_root
    bin_name = config.project.bin_name
    if not bin_name:
        logger.error("No bin_name in config!")
        sys.exit(1)
    bin_name = bin_name.lower()
    bin_path = project_path / "build" / flash_profile / "bin" / f"{bin_name}.bin"

    flash_start = hex(config.get_stm32_config().flash_start)

    if not bin_path.exists():
        logger.error("%s does not exist", bin_path)
        logger.info("Did you run scargo build --profile %s", flash_profile)
    elif not flash_start:
        logger.error("Flash start address not found in lock file")
        logger.info("Define flash-start in scargo.toml under stm32 section")
    else:
        subprocess.check_call(["st-flash", "erase"])
        subprocess.check_call(
            ["st-flash", "--reset", "write", str(bin_path), flash_start]
        )



# #
# @copyright Copyright (C) 2023 SpyroSoft Solutions S.A. All rights reserved.
# #
from pathlib import Path
from typing import List

COPYRIGHT = """# #
# Copyright (C) 2022 Spyrosoft Solutions. All rights reserved.
# #\n\n"""


def add_subdirs_to_cmake(mock_root: Path) -> None:
    """This function takes as input a path to mocks and a path to root.
    Then it creates a list of root directories where the index of "src"
    directory is found. The "start_path" is created which is a path of
    first "src_idx" directories. This path should be looped over to add
    subdirectories into CMakeLists.txt."""

    mock_root = mock_root.resolve()

    if not mock_root.exists():
        mock_root = mock_root.parent

    while True:
        subdirectories = [path for path in mock_root.iterdir() if path.is_dir()]
        if subdirectories:
            break
        mock_root = mock_root.parent

    try:
        update_cmake_lists(mock_root, subdirectories)
    except OSError:
        pass


def update_cmake_lists(mock_root: Path, subdirectories: List[Path]) -> None:
    # if CMakelists doesn't exist create it in mock_root
    cmake_list_path = mock_root / "CMakeLists.txt"
    if not cmake_list_path.exists():
        with cmake_list_path.open("w", encoding="utf-8") as file:
            file.write(COPYRIGHT)
    # check if all subidrectories exists in CMakeLists
    # if not add new ones
    with cmake_list_path.open("r+", encoding="utf-8") as file:
        cmake_text = file.read()
        for subdir in subdirectories:
            if str(subdir.relative_to(mock_root)) not in cmake_text:
                file.write(f"\nadd_subdirectory({subdir})")



# #
# @copyright Copyright (C) 2023 SpyroSoft Solutions S.A. All rights reserved.
# #

import re
from pathlib import Path
from typing import List, Sequence

from scargo.config import Config
from scargo.file_generators.base_gen import create_file_from_template
from scargo.file_generators.mock_utils.cmake_utils import add_subdirs_to_cmake

HEADER_EXTENSIONS = (".h", ".hpp")
SRC_EXTENSIONS = (".c", ".cpp")


class HeaderDescriptor:
    """
    This class is a container for names, includes, classes and namespaces
    which are parsed from the header file. This information is then passed
    to generate unit tests.
    """

    def __init__(
        self, name: str, includes: List[str], classes: List[str], namespaces: List[str]
    ) -> None:
        self.name = name
        self.includes = includes
        self.classes = classes
        self.namespaces = namespaces


class _UnitTestsGen:
    def __init__(self, config: Config):
        self._config = config
        self._project_path = config.project_root
        self._ut_dir = self._project_path / "tests/ut"

    def generate_tests(self, input_path: Path, overwrite: bool) -> None:
        """Generates unit test files and corresponding cmake file

        :param Path input_path: Path to src file or src directory
        :param bool overwrite: overwrite files if exist
        """
        if input_path.is_file():
            ut_path = self._get_unit_test_path(input_path)
            self._generate_unit_test(input_path, ut_path, overwrite)
            self._generate_cmake(input_path.parent, ut_path.parent)

        elif input_path.is_dir():
            headers = self._get_paths_with_ext(input_path, HEADER_EXTENSIONS)
            ut_path = None
            for hdr in headers:
                ut_path = self._get_unit_test_path(hdr)
                self._generate_unit_test(hdr, ut_path, overwrite)
            if ut_path:
                self._generate_cmake(input_path, ut_path.parent)

    def _generate_unit_test(
        self, input_file_path: Path, output_file_path: Path, overwrite: bool
    ) -> None:
        """Generates unit test source file

        :param Path input_file_path: Path to src file
        :param Path output_file_path: Path to unit test file
        :param bool overwrite: overwrite if exists
        """
        header_descriptor = self._parse_header_file(input_file_path)
        create_file_from_template(
            "ut/ut.cpp.j2",
            output_file_path,
            overwrite=overwrite,
            template_params={"header": header_descriptor},
            config=self._config,
        )

    def _generate_cmake(self, src_dir_path: Path, ut_dir_path: Path) -> None:
        """Generate CMakeLists for unit tests

        :param Path src_dir_path: Source directory for which tests are being generated
        :param Path ut_dir_path: Directory of generated unit tests
        """
        add_subdirs_to_cmake(ut_dir_path.relative_to(self._project_path))

        ut_name = self._get_cmake_tests_name(ut_dir_path)
        ut_files = [
            p.name for p in self._get_paths_with_ext(ut_dir_path, SRC_EXTENSIONS)
        ]

        # Exclude main from srcs to test
        main_cpp = (
            f"{self._config.project.bin_name}.cpp"
            if self._config.project.bin_name
            else None
        )

        src_files = [
            p.absolute().relative_to(self._project_path.absolute())
            for p in self._get_paths_with_ext(src_dir_path, SRC_EXTENSIONS)
            if p.name != main_cpp
        ]

        create_file_from_template(
            "ut/CMakeLists.txt.j2",
            ut_dir_path / "CMakeLists.txt",
            overwrite=True,
            template_params={
                "src_files": src_files,
                "utest_name": ut_name,
                "ut_files": ut_files,
            },
            config=self._config,
        )

    def _get_unit_test_path(self, input_src_path: Path) -> Path:
        """Return output path for unit test

        :param Path input_src_path: Source path
        :return Path: output path for unit test
        """
        input_src_path = input_src_path.absolute()
        relative_to_src = input_src_path.relative_to(
            self._project_path.absolute() / self._config.project.target.source_dir
        )
        return Path(self._ut_dir, relative_to_src).with_name(
            f"ut_{input_src_path.stem}.cpp"
        )

    def _get_cmake_tests_name(self, test_dir_path: Path) -> str:
        """Get tests name for cmake

        :param Path test_dir_path: Path to test dir
        :return str: tests name for cmake
        """
        test_dir_path = test_dir_path.absolute()
        relative_path = test_dir_path.relative_to(self._project_path.absolute())
        return "_".join(relative_path.parts)

    @staticmethod
    def _get_namespace(line: str) -> str:
        if line.startswith("namespace"):
            namespace = line.split(" ")[1]
        elif line.startswith("using namespace"):
            namespace = line.split(" ")[2]
        else:
            raise ValueError(f"No 'namespace' found in line: '{line}'")

        last_char = namespace[-1]
        if last_char in ("{", ";"):
            return namespace[0:-1]

        return namespace

    @staticmethod
    def _parse_header_file(header_path: Path) -> HeaderDescriptor:
        namespaces = []
        classes = []
        includes = []

        # open header file to parse the header name and class name
        with open(header_path, encoding="utf-8") as header:
            for line in header:
                line = line.strip()
                if line.startswith("#include"):
                    includes.append(line.split(" ")[1])
                elif line.startswith("namespace") or line.startswith("using namespace"):
                    namespaces.append(_UnitTestsGen._get_namespace(line))
                elif line.startswith("class"):
                    class_regex = re.compile(r"^([a-z]+)(\s[a-zA-Z\d]+)")
                    temp = class_regex.search(line)
                    if temp:
                        classes.append(temp.group(2).lstrip())
                elif line.startswith('extern "C"'):
                    class_name = "".join([w.capitalize() for w in header_path.stem])
                    classes.append(class_name)

        return HeaderDescriptor(str(header_path), namespaces, classes, includes)

    @staticmethod
    def _get_paths_with_ext(workdir: Path, extensions: Sequence[str]) -> List[Path]:
        return [child for child in workdir.iterdir() if child.suffix in extensions]


def generate_ut(input_path: Path, config: Config, force: bool = False) -> None:
    ut_gen = _UnitTestsGen(config)
    ut_gen.generate_tests(input_path, force)
